/********************************************************************************/
/* matmul_strassen.h:                                                           */
/* Copyright (C) 2014 Tomonori Kouya                                            */
/*                                                                              */
/* This program is free software: you can redistribute it and/or modify it      */
/* under the terms of the GNU Lesser General Public License as published by the */
/* Free Software Foundation, either version 3 of the License or any later       */
/* version.                                                                     */
/*                                                                              */
/* This program is distributed in the hope that it will be useful, but WITHOUT  */
/* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or        */
/* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License */
/* for more details.                                                            */
/*                                                                              */
/* You should have received a copy of the GNU Lesser General Public License     */
/* along with this program.  If not, see <http://www.gnu.org/licenses/>.        */
/*                                                                              */
/********************************************************************************/
#include <stdio.h>
#include <math.h>

#include "bnc.h"

#ifdef USE_IMKL
	#include "mkl.h"
	#include "mkl_cblas.h" // for Intel Math Kernel Library
#endif

// define _BNC_MATMUL_STRASSEN_H
#ifndef _BNC_MATMUL_STRASSEN_H
#define _BNC_MATMUL_STRASSEN_H

#ifndef _BNC_DEFAULT_MIN_DIM_STRASSEN
//	#define _BNC_DEFAULT_MIN_DIM_STRASSEN 2
//	#define _BNC_DEFAULT_MIN_DIM_STRASSEN 4
//	#define _BNC_DEFAULT_MIN_DIM_STRASSEN 8
//	#define _BNC_DEFAULT_MIN_DIM_STRASSEN 16
	#define _BNC_DEFAULT_MIN_DIM_STRASSEN 32
//	#define _BNC_DEFAULT_MIN_DIM_STRASSEN 64
#endif // _BNC_DEFAULT_MIN_DIM_STRASSEN

// variables to count the number of arithmetic
// double precision
long int num_addsub_mul_dmatrix_strassen;	// addition and subtraction
long int num_mul_mul_dmatrix_strassen;		// multiplication

// clear counter
void reset_num_mul_dmatrix_strassen(void);

// print counters
void print_num_mul_dmatrix_strassen(long int *num_addsub, long int *num_mul);

// multiple precision
long int num_addsub_mul_mpfmatrix_strassen;	// addition and subtraction
long int num_mul_mul_mpfmatrix_strassen;		// multiplication

// clear counter
void reset_num_mul_mpfmatrix_strassen(void);

// get counters
void get_num_mul_mpfmatrix_strassen(long int *num_addsub, long int *num_mul);

// print counters
void print_num_mul_mpfmatrix_strassen(long int *num_addsub, long int *num_mul);

/* c = a * b */
void mul_dmatrix_simple(DMatrix c, DMatrix a, DMatrix b);

// partial add
// *_index[0] = start_row_index
// *_index[1] = end_row_index
// *_index[2] = start_col_index
// *_index[3] = end_col_index
void add_dmatrix_partial(DMatrix ret, long int ret_index[4], DMatrix mat_a, long int mat_a_index[4], DMatrix mat_b, long int mat_b_index[4]);

// partial sub
// *_index[0] = start_row_index
// *_index[1] = end_row_index
// *_index[2] = start_col_index
// *_index[3] = end_col_index
void sub_dmatrix_partial(DMatrix ret, long int ret_index[4], DMatrix mat_a, long int mat_a_index[4], DMatrix mat_b, long int mat_b_index[4]);

// partial set
// *_index[0] = start_row_index
// *_index[1] = end_row_index
// *_index[2] = start_col_index
// *_index[3] = end_col_index
void subst_dmatrix_partial(DMatrix ret, long int ret_index[4], DMatrix mat_a, long int mat_a_index[4]);

// partial set
// *_index[0] = start_row_index
// *_index[1] = end_row_index
// *_index[2] = start_col_index
// *_index[3] = end_col_index
void subst_dmatrix_partial_checked(DMatrix ret, long int ret_index[4], DMatrix mat_a, long int mat_a_index[4]);

// partial set
// *_index[0] = start_row_index
// *_index[1] = end_row_index
// *_index[2] = start_col_index
// *_index[3] = end_col_index
void neg_dmatrix_partial(DMatrix ret, long int ret_index[4], DMatrix mat_a, long int mat_a_index[4]);

// Padding to 2-powered dimensional matrix
DMatrix init_static_padding_dmatrix_strassen(DMatrix orig_mat);

// Padding to even dimensional matrix
DMatrix init_dynamic_padding_dmatrix_strassen(DMatrix orig_mat);

// Strassen's Algorithm with static padding
void mul_dmatrix_strassen_odd_padding(DMatrix ret, DMatrix mat_a, DMatrix mat_b, long int min_dim);

// Strassen's Algorithm with dynamic peeling
void mul_dmatrix_strassen(DMatrix ret, DMatrix mat_a, DMatrix mat_b, long int min_dim);

// Strassen's Algorithm
void mul_dmatrix_strassen_odd_peeling(DMatrix ret, DMatrix mat_a, DMatrix mat_b, long int min_dim);

// Strassen's Algorithm
void mul_dmatrix_strassen_even(DMatrix ret, DMatrix mat_a, DMatrix mat_b, long int min_dim);

// Winograd Variant of Strassen's Algorithm
void mul_dmatrix_winograd_even(DMatrix ret, DMatrix mat_a, DMatrix mat_b, long int min_dim);

// Block matrix multiplicaiton
void mul_dmatrix_block(DMatrix ret, DMatrix mat_a, DMatrix mat_b, long int min_dim);

// Computattion of Inverse Matrix by using Strassen's Algorithm
void inv_dmatrix_strassen_even(DMatrix ret, DMatrix mat_a, long int min_dim);

// MPF & MPFR
#ifdef USE_GMP

// partial add
// *_index[0] = start_row_index
// *_index[1] = end_row_index
// *_index[2] = start_col_index
// *_index[3] = end_col_index
void add_mpfmatrix_partial(MPFMatrix ret, long int ret_index[4], MPFMatrix mat_a, long int mat_a_index[4], MPFMatrix mat_b, long int mat_b_index[4]);

// partial add
// *_index[0] = start_row_index
// *_index[1] = end_row_index
// *_index[2] = start_col_index
// *_index[3] = end_col_index
void sub_mpfmatrix_partial(MPFMatrix ret, long int ret_index[4], MPFMatrix mat_a, long int mat_a_index[4], MPFMatrix mat_b, long int mat_b_index[4]);

// partial set
// *_index[0] = start_row_index
// *_index[1] = end_row_index
// *_index[2] = start_col_index
// *_index[3] = end_col_index
void subst_mpfmatrix_partial(MPFMatrix ret, long int ret_index[4], MPFMatrix mat_a, long int mat_a_index[4]);

// partial set
// *_index[0] = start_row_index
// *_index[1] = end_row_index
// *_index[2] = start_col_index
// *_index[3] = end_col_index
void neg_mpfmatrix_partial(MPFMatrix ret, long int ret_index[4], MPFMatrix mat_a, long int mat_a_index[4]);

/* c = a * b */
void mul_mpfmatrix_simple(MPFMatrix c, MPFMatrix a, MPFMatrix b);

// Block matrix multiplicaiton
void mul_mpfmatrix_block(MPFMatrix ret, MPFMatrix mat_a, MPFMatrix mat_b, long int min_dim);

// Padding to 2-powered dimensional matrix
MPFMatrix init_static_padding_mpfmatrix_strassen(MPFMatrix orig_mat);

// Padding to even dimensional matrix
MPFMatrix init_dynamic_padding_mpfmatrix_strassen(MPFMatrix orig_mat);

// Strassen's Algorithm with static or dynamic padding
void mul_mpfmatrix_strassen_odd_padding(MPFMatrix ret, MPFMatrix mat_a, MPFMatrix mat_b, long int min_dim);

// Fit dimension to be multiple of min_dim
void mul_mpfmatrix_strassen(MPFMatrix ret, MPFMatrix mat_a, MPFMatrix mat_b, long int min_dim);

// Strassen's Algorithm (even dimension)
void mul_mpfmatrix_strassen_even(MPFMatrix ret, MPFMatrix mat_a, MPFMatrix mat_b, long int min_dim);

// Winograd Variant of Strassen's Algorithm
void mul_mpfmatrix_winograd_even(MPFMatrix ret, MPFMatrix mat_a, MPFMatrix mat_b, long int min_dim);

// Strassen's Algorithm with dynamic peeling
void mul_mpfmatrix_strassen_odd_peeling(MPFMatrix ret, MPFMatrix mat_a, MPFMatrix mat_b, long int min_dim);

// Computation of Inverse Matrix by using Strassen's Algorithm
void inv_mpfmatrix_strassen_even(MPFMatrix ret, MPFMatrix mat_a, long int min_dim);

#endif // USE_GMP

#endif // define _BNC_MATMUL_STRASSEN_H
